"""Basic smoke tests."""

import json

from pathlib import Path

import pytest

from pip4a.cli import main
from pip4a.output import Output
from pip4a.utils import TermFeatures, subprocess_run


def test_venv(
    capsys: pytest.CaptureFixture[str],
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Basic smoke test."""
    # disable color for json output
    term_features = TermFeatures(color=False, links=False)
    output = Output(
        log_file=f"/{tmp_path}/pip4a.log",
        log_level="INFO",
        log_append="false",
        term_features=term_features,
        verbosity=0,
    )
    command = (
        "git clone https://github.com/ansible-collections/cisco.nxos.git"
        f" {tmp_path/ 'cisco.nxos'}"
    )
    subprocess_run(
        command=command,
        verbose=True,
        msg="",
        output=output,
    )
    monkeypatch.chdir(tmp_path)

    monkeypatch.setattr(
        "sys.argv",
        ["pip4a", "install", str(tmp_path / "cisco.nxos"), "--venv=venv", "--no-ansi"],
    )
    with pytest.raises(SystemExit):
        main()
    string = "Installed collections include: ansible.netcommon, ansible.utils,"
    captured = capsys.readouterr()

    assert string in captured.out

    monkeypatch.setattr(
        "sys.argv",
        ["pip4a", "list", "--venv=venv"],
    )
    with pytest.raises(SystemExit):
        main()
    captured = capsys.readouterr()
    assert "cisco.nxos" in captured.out
    assert "ansible.netcommon" in captured.out
    assert "ansible.utils" in captured.out
    assert "unknown" not in captured.out

    monkeypatch.setattr(
        "sys.argv",
        ["pip4a", "uninstall", "ansible.utils", "--venv=venv"],
    )
    with pytest.raises(SystemExit):
        main()
    captured = capsys.readouterr()
    assert "Removed ansible.utils" in captured.out

    monkeypatch.setattr("sys.argv", ["pip4a", "inspect", "--venv=venv", "--no-ansi"])
    with pytest.raises(SystemExit):
        main()
    captured = capsys.readouterr()
    captured_json = json.loads(captured.out)
    assert "cisco.nxos" in captured_json
    assert "ansible.netcommon" in captured_json
    assert "ansible.utils" not in captured_json

    command = f"{tmp_path / 'venv' / 'bin' / 'python'} -m pip uninstall xmltodict -y"
    subprocess_run(command=command, verbose=True, msg="", output=output)

    monkeypatch.setattr("sys.argv", ["pip4a", "check", "--venv=venv"])
    with pytest.raises(SystemExit):
        main()
    captured = capsys.readouterr()
    assert "Missing python dependencies: xmltodict" in captured.err


def test_non_local(
    capsys: pytest.CaptureFixture[str],
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Install non-local collection."""
    monkeypatch.setattr(
        "sys.argv",
        ["pip4a", "install", "ansible.scm", f"--venv={tmp_path / 'venv'}"],
    )
    with pytest.raises(SystemExit):
        main()
    string = "Installed collections include: ansible.scm and ansible.utils"
    captured = capsys.readouterr()
    assert string in captured.out
    monkeypatch.setattr(
        "sys.argv",
        ["pip4a", "tree", f"--venv={tmp_path / 'venv'}"],
    )
    with pytest.raises(SystemExit):
        main()
    captured = capsys.readouterr()
    with pytest.raises(SystemExit):
        main()
    captured = capsys.readouterr()
    string = "ansible.scm\n└──ansible.utils\n\n"
    assert string == captured.out


def test_requirements(
    capsys: pytest.CaptureFixture[str],
    tmp_path: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Install non-local collection."""
    requirements = Path(__file__).parent.parent / "fixtures" / "requirements.yml"
    monkeypatch.setattr(
        "sys.argv",
        ["pip4a", "install", f"--venv={tmp_path / 'venv'}", "-r", str(requirements)],
    )
    with pytest.raises(SystemExit):
        main()
    string = "Installed collections include: ansible.netcommon, ansible.scm,"
    captured = capsys.readouterr()
    assert string in captured.out
    monkeypatch.setattr(
        "sys.argv",
        ["pip4a", "uninstall", f"--venv={tmp_path / 'venv'}", "-r", str(requirements)],
    )
    with pytest.raises(SystemExit):
        main()
    captured = capsys.readouterr()
    string = "Removed ansible.netcommon"
    assert string in captured.out
    string = "Removed ansible.scm"
    assert string in captured.out

    monkeypatch.setattr(
        "sys.argv",
        ["pip4a", "list", f"--venv={tmp_path / 'venv'}"],
    )
    with pytest.raises(SystemExit):
        main()
    captured = capsys.readouterr()
    assert "ansible.netcommon" not in captured.out
    assert "ansible.scm" not in captured.out
    assert "ansible.utils" in captured.out



"""An ascii tree generator."""
from __future__ import annotations

from typing import Union

from .utils import Ansi, TermFeatures, term_link


ScalarVal = Union[bool, str, float, int, None]
JSONVal = Union[ScalarVal, list["JSONVal"], dict[str, "JSONVal"]]


class Tree:  # pylint: disable=R0902
    """Renderer for the tree."""

    PIPE = "│"
    ELBOW = "└──"
    TEE = "├──"
    PIPE_PREFIX = "│  "
    SPACE_PREFIX = "   "

    def __init__(
        self: Tree,
        obj: JSONVal,
        term_features: TermFeatures,
    ) -> None:
        """Initialize the renderer."""
        self.obj = obj
        self._lines: list[str] = []
        self.blue: list[ScalarVal] = []
        self.bold: list[ScalarVal] = []
        self.cyan: list[ScalarVal] = []
        self.green: list[ScalarVal] = []
        self.italic: list[ScalarVal] = []
        self.magenta: list[ScalarVal] = []
        self.red: list[ScalarVal] = []
        self.reversed: list[ScalarVal] = []
        self.underline: list[ScalarVal] = []
        self.white: list[ScalarVal] = []
        self.yellow: list[ScalarVal] = []
        self.links: dict[str, str] = {}
        self.term_features = term_features

    def in_color(self: Tree, val: ScalarVal) -> str:
        """Colorize the string.

        Args:
            val: The thing to colorize

        Returns:
            The colorized string
        """
        if not self.term_features.color:
            return str(val)

        ansis = (
            "blue",
            "bold",
            "cyan",
            "green",
            "italic",
            "magenta",
            "red",
            "reversed",
            "underline",
            "white",
            "yellow",
        )
        start = ""
        val_str = str(val)
        for ansi in ansis:
            matches = getattr(self, ansi)
            if val_str in [str(match) for match in matches]:
                start += getattr(Ansi, ansi.upper())
        if val_str in self.links:
            val_str = term_link(
                uri=self.links[val_str],
                term_features=self.term_features,
                label=val_str,
            )

        if start:
            return f"{start}{val_str}{Ansi.RESET}"

        return val_str

    @staticmethod
    def is_scalar(obj: JSONVal) -> bool:
        """Check if the object is a scalar."""
        return isinstance(obj, (str, int, float, bool)) or obj is None

    def _print_tree(  # noqa: C901, PLR0913, PLR0912
        self: Tree,
        obj: JSONVal,
        is_last: bool,  # noqa: FBT001
        is_root: bool,  # noqa: FBT001
        was_list: bool,  # noqa: FBT001
        prefix: str = "",
    ) -> None:
        """Print the tree.

        Args:
            obj: The object to print
            is_last: Whether the object is the last in the list | dict
            is_root: Whether the object is the root of the tree
            was_list: Whether the object was a list
            prefix: The prefix to use

        Raises:
            TypeError: If the object is not a dict, list, or scalar
        """
        # pylint: disable=R0914
        if isinstance(obj, dict):
            for i, (key, value) in enumerate(obj.items()):
                is_last = i == len(obj) - 1
                key_repr = f"{Ansi.ITALIC}{key}{Ansi.RESET}" if was_list else key
                if is_root:
                    decorator = ""
                elif is_last:
                    decorator = self.ELBOW
                else:
                    decorator = self.TEE
                self.append(f"{prefix}{decorator}{self.in_color(key_repr)}")

                if is_root:
                    prefix_rev = prefix
                elif is_last:
                    prefix_rev = prefix + self.SPACE_PREFIX
                else:
                    prefix_rev = prefix + self.PIPE_PREFIX
                self._print_tree(
                    obj=value,
                    prefix=prefix_rev,
                    is_last=self.is_scalar(value),
                    is_root=False,
                    was_list=False,
                )

        elif isinstance(obj, list):
            is_complex = any(isinstance(item, (dict, list)) for item in obj)
            is_long = len(obj) > 1
            if is_complex and is_long:
                repr_obj = {str(i): item for i, item in enumerate(obj)}
                self._print_tree(
                    obj=repr_obj,
                    prefix=prefix,
                    is_last=is_last,
                    is_root=False,
                    was_list=True,
                )
            else:
                for i, item in enumerate(obj):
                    is_last = i == len(obj) - 1
                    self._print_tree(
                        obj=item,
                        prefix=prefix,
                        is_last=is_last,
                        is_root=False,
                        was_list=False,
                    )

        elif self.is_scalar(obj):
            self.append(
                f"{prefix}{self.ELBOW if is_last else self.TEE}{self.in_color(obj)}",
            )
        else:
            err = f"Invalid type {type(obj)}"
            raise TypeError(err)

    def append(self: Tree, string: str) -> None:
        """Append a line to the output."""
        self._lines.append(string)

    def render(self: Tree) -> str:
        """Render the root of the tree."""
        # if not isinstance(self.obj, dict):
        self._print_tree(self.obj, is_last=False, is_root=True, was_list=False)
        return "\n".join(self._lines) + "\n"



"""Utility functions."""

from __future__ import annotations

import itertools
import json
import logging
import subprocess
import sys
import threading
import time

from dataclasses import dataclass
from typing import TYPE_CHECKING, Union

import subprocess_tee
import yaml


if TYPE_CHECKING:
    from pathlib import Path
    from types import TracebackType

    from .config import Config
    from .output import Output

from typing import Any


logger = logging.getLogger(__name__)


ScalarVal = Union[bool, str, float, int, None]
JSONVal = Union[ScalarVal, list["JSONVal"], dict[str, "JSONVal"]]


@dataclass
class TermFeatures:
    """Terminal features."""

    color: bool
    links: bool

    def any_enabled(self: TermFeatures) -> bool:
        """Return True if any features are enabled."""
        return any((self.color, self.links))


def term_link(uri: str, term_features: TermFeatures, label: str) -> str:
    """Return a link.

    Args:
        uri: The URI to link to
        term_features: The terminal features to enable
        label: The label to use for the link
    Returns:
        The link
    """
    if not term_features.links:
        return label

    parameters = ""

    # OSC 8 ; params ; URI ST <name> OSC 8 ;; ST
    escape_mask = "\x1b]8;{};{}\x1b\\{}\x1b]8;;\x1b\\"
    link_str = escape_mask.format(parameters, uri, label)
    if not term_features.color:
        return link_str
    return f"{Ansi.BLUE}{link_str}{Ansi.RESET}"


class Ansi:
    """ANSI escape codes."""

    BLUE = "\x1B[34m"
    BOLD = "\x1B[1m"
    CYAN = "\x1B[36m"
    GREEN = "\x1B[32m"
    ITALIC = "\x1B[3m"
    MAGENTA = "\x1B[35m"
    RED = "\x1B[31m"
    RESET = "\x1B[0m"
    REVERSED = "\x1B[7m"
    UNDERLINE = "\x1B[4m"
    WHITE = "\x1B[37m"
    YELLOW = "\x1B[33m"
    GREY = "\x1B[90m"


def subprocess_run(  # noqa: PLR0913
    command: str,
    verbose: int,
    msg: str,
    output: Output,
    cwd: Path | None = None,
    env: dict[str, str] | None = None,
) -> subprocess.CompletedProcess[str]:
    """Run a subprocess command."""
    cmd = f"Running command: {command}"
    output.debug(cmd)
    log_level = logging.ERROR - (verbose * 10)
    if log_level == logging.DEBUG:
        return subprocess_tee.run(
            command,
            check=True,
            cwd=cwd,
            env=env,
            shell=True,  # noqa: S604
            text=True,
        )
    with Spinner(message=msg, term_features=output.term_features):
        return subprocess.run(
            command,
            check=True,
            cwd=cwd,
            env=env,
            shell=True,  # noqa: S602
            capture_output=True,
            text=True,
        )


def oxford_join(words: list[str]) -> str:
    """Join a list of words with commas and an oxford comma.

    :param words: A list of words to join
    :return: A string of words joined with commas and an oxford comma
    """
    words.sort()
    if not words:
        return ""
    if len(words) == 1:
        return words[0]
    if len(words) == 2:  # noqa: PLR2004
        return " and ".join(words)
    return ", ".join(words[:-1]) + ", and " + words[-1]


def opt_deps_to_files(collection_path: Path, dep_str: str) -> list[Path]:
    """Convert a string of optional dependencies to a list of files.

    :param dep_str: A string of optional dependencies
    :return: A list of files
    """
    deps = dep_str.split(",")
    files = []
    for dep in deps:
        _dep = dep.strip()
        variant1 = collection_path / f"{_dep}-requirements.txt"
        if variant1.exists():
            files.append(variant1)
            continue
        variant2 = collection_path / f"requirements-{_dep}.txt"
        if variant2.exists():
            files.append(variant2)
            continue
        msg = (
            f"Failed to find optional dependency file for '{_dep}'."
            f" Checked for '{variant1.name}' and '{variant2.name}'. Skipping."
        )
        logger.error(msg)
    return files


def sort_dict(item: dict[str, Any]) -> dict[str, Any]:
    """Recursively sort a dictionary.

    Args:
        item: The dictionary to sort.

    Returns:
        The sorted dictionary.
    """
    return {
        k: sort_dict(v) if isinstance(v, dict) else v for k, v in sorted(item.items())
    }


def collect_manifests(  # noqa: C901
    target: Path,
    venv_cache_dir: Path,
) -> dict[str, dict[str, JSONVal]]:
    # pylint: disable=too-many-locals
    """Collect manifests from a target directory.

    Args:
        target: The target directory to collect manifests from.
        venv_cache_dir: The directory to look for manifests in.

    Returns:
        A dictionary of manifests.
    """
    collections = {}
    for namespace_dir in target.iterdir():
        if not namespace_dir.is_dir():
            continue

        for name_dir in namespace_dir.iterdir():
            if not name_dir.is_dir():
                continue
            manifest = name_dir / "MANIFEST.json"
            if not manifest.exists():
                manifest = (
                    venv_cache_dir
                    / f"{namespace_dir.name}.{name_dir.name}"
                    / "MANIFEST.json"
                )
            if not manifest.exists():
                msg = f"Manifest not found for {namespace_dir.name}.{name_dir.name}"
                logger.debug(msg)
                continue
            with manifest.open() as manifest_file:
                manifest_json = json.load(manifest_file)

            cname = f"{namespace_dir.name}.{name_dir.name}"

            collections[cname] = manifest_json
            c_info = collections[cname].get("collection_info", {})
            if not c_info:
                collections[cname]["collection_info"] = {}
                c_info = collections[cname]["collection_info"]
            c_info["requirements"] = {"python": {}, "system": []}

            python_requirements = c_info["requirements"]["python"]
            system_requirements = c_info["requirements"]["system"]

            for file in name_dir.iterdir():
                if not file.is_file():
                    continue
                if not file.name.endswith(".txt"):
                    continue
                if "requirements" in file.name:
                    with file.open() as requirements_file:
                        requirements = requirements_file.read().splitlines()
                        python_requirements[file.stem] = requirements
                if file.stem == "bindep":
                    with file.open() as requirements_file:
                        requirements = requirements_file.read().splitlines()
                        system_requirements.extend(requirements)

    return sort_dict(collections)


def builder_introspect(config: Config, output: Output) -> None:
    """Introspect a collection.

    Args:
        config: The configuration object.
        output: The output object.
    """
    command = (
        f"ansible-builder introspect {config.site_pkg_path}"
        f" --write-pip {config.discovered_python_reqs}"
        f" --write-bindep {config.discovered_bindep_reqs}"
        " --sanitize"
    )
    if (
        hasattr(config.args, "collection_specifier")
        and hasattr(config, "collection")
        and config.collection.opt_deps
        and config.collection.path
    ):
        dep_paths = opt_deps_to_files(
            collection_path=config.collection.path,
            dep_str=config.collection.opt_deps,
        )
        for dep_path in dep_paths:
            command += f" --user-pip {dep_path}"
    msg = f"Writing discovered python requirements to: {config.discovered_python_reqs}"
    logger.debug(msg)
    msg = f"Writing discovered system requirements to: {config.discovered_bindep_reqs}"
    logger.debug(msg)
    work = "Persisting requirements to file system"
    try:
        subprocess_run(
            command=command,
            verbose=config.args.verbose,
            msg=work,
            output=output,
        )
    except subprocess.CalledProcessError as exc:
        err = f"Failed to discover requirements: {exc} {exc.stderr}"
        logger.critical(err)

    if not config.discovered_python_reqs.exists():
        config.discovered_python_reqs.touch()

    if not config.discovered_bindep_reqs.exists():
        config.discovered_bindep_reqs.touch()


def collections_from_requirements(file: Path) -> list[dict[str, str]]:
    """Build a list of collections from a requirements file."""
    collections = []
    try:
        with file.open() as requirements_file:
            requirements = yaml.safe_load(requirements_file)
    except yaml.YAMLError as exc:
        err = f"Failed to load yaml file: {exc}"
        logger.critical(err)

    for requirement in requirements["collections"]:
        if isinstance(requirement, str):
            collections.append({"name": requirement})
        elif isinstance(requirement, dict):
            collections.append(requirement)
    return collections


def collections_meta(config: Config) -> dict[str, dict[str, Any]]:
    """Collect metadata about installed collections.

    Args:
        config: The configuration object.
    Returns:
        A dictionary of metadata about installed collections.
    """
    all_info_dirs = [
        entry
        for entry in config.site_pkg_collections_path.iterdir()
        if entry.name.endswith(".info")
    ]

    collections = {}
    for namespace_dir in config.site_pkg_collections_path.iterdir():
        if not namespace_dir.is_dir():
            continue

        for name_dir in namespace_dir.iterdir():
            if not name_dir.is_dir():
                continue
            some_info_dirs = [
                info_dir
                for info_dir in all_info_dirs
                if f"{namespace_dir.name}.{name_dir.name}" in info_dir.name
            ]
            file = None
            if some_info_dirs:
                file = some_info_dirs[0] / "GALAXY.yml"
                editable_location = ""

            elif (name_dir / "galaxy.yml").exists():
                file = name_dir / "galaxy.yml"
                editable_location = (
                    str(name_dir.resolve()) if name_dir.is_symlink() else ""
                )

            if file:
                with file.open() as info_file:
                    info = yaml.safe_load(info_file)
                    collections[f"{namespace_dir.name}.{name_dir.name}"] = {
                        "version": info.get("version", "unknown"),
                        "editable_location": editable_location,
                        "dependencies": info.get("dependencies", []),
                    }
            else:
                collections[f"{namespace_dir.name}.{name_dir.name}"] = {
                    "version": "unknown",
                    "editable_location": "",
                    "dependencies": [],
                }
    return collections


class Spinner:  # pylint: disable=too-many-instance-attributes
    """A spinner."""

    def __init__(
        self: Spinner,
        message: str,
        term_features: TermFeatures,
        delay: float = 0.1,
    ) -> None:
        """Initialize the spinner.

        Args:
            message: The message to display
            term_features: Terminal features
            delay: The delay between characters
        """
        self.spinner = itertools.cycle(("|", "/", "-", "\\", "|", "/", "-"))
        self.delay = delay
        self.busy = False
        self.spinner_visible = False
        self._term_features = term_features
        self._screen_lock = threading.Lock()
        self._start_time: float | None = None
        self.thread: threading.Thread
        self.msg: str = message.rstrip(".").rstrip(":").rstrip()

    def write_next(self: Spinner) -> None:
        """Write the next char."""
        with self._screen_lock:
            if not self.spinner_visible:
                if self._term_features.color:
                    char = f"{Ansi.GREY}{next(self.spinner)}{Ansi.RESET}"
                else:
                    char = next(self.spinner)
                sys.stdout.write(char)
                self.spinner_visible = True
                sys.stdout.flush()

    def remove_spinner(
        self: Spinner,
        cleanup: bool = False,  # noqa: FBT001,FBT002
    ) -> None:
        """Remove the spinner.

        https://github.com/Tagar/stuff/blob/master/spinner.py

        Args:
            cleanup: Should we cleanup after the spinner
        """
        with self._screen_lock:
            if self.spinner_visible:
                sys.stdout.write("\b")
                self.spinner_visible = False
                if cleanup:
                    sys.stdout.write(" ")  # overwrite spinner with blank
                    sys.stdout.write("\r")  # move to next line
                    sys.stdout.write("\033[K")  # clear line
                sys.stdout.flush()

    def spinner_task(self: Spinner) -> None:
        """Spin the spinner."""
        while self.busy:
            self.write_next()
            time.sleep(self.delay)
            self.remove_spinner()

    def __enter__(self: Spinner) -> None:
        """Enter the context handler."""
        # set the start time
        self._start_time = time.time()
        if not self._term_features.any_enabled():
            return
        if self._term_features.color:
            sys.stdout.write(f"{Ansi.GREY}{self.msg}:{Ansi.RESET} ")
        else:
            sys.stdout.write(f"{self.msg}: ")
        # hide the cursor
        sys.stdout.write("\033[?25l")
        if self._term_features.any_enabled():
            self.busy = True
            self.thread = threading.Thread(target=self.spinner_task)
            self.thread.start()

    def __exit__(
        self: Spinner,
        typ: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        """Exit the context handler.

        Args:
            typ: The exception
            exc: The exception value
            tb: The traceback


        """
        # delay if less than n seconds has elapsed
        if not self._term_features.any_enabled():
            return
        min_show_time = 0.5
        if self._start_time:
            elapsed = time.time() - self._start_time
            if elapsed < min_show_time:
                time.sleep(min_show_time - elapsed)
        if self._term_features.any_enabled():
            self.busy = False
            self.remove_spinner(cleanup=True)
        else:
            sys.stdout.write("\r")
        # show the cursor
        sys.stdout.write("\033[?25h")



"""Output functionality."""

from __future__ import annotations

import logging
import os
import sys
import textwrap

from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar


if TYPE_CHECKING:
    from .utils import TermFeatures


T = TypeVar("T", bound="Level")


class Color:
    """Color constants."""

    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    GREY = "\033[90m"  # Bright black?
    BRIGHT_RED = "\033[91m"
    BRIGHT_GREEN = "\033[92m"
    BRIGHT_YELLOW = "\033[93m"
    BRIGHT_BLUE = "\033[94m"
    BRIGHT_MAGENTA = "\033[95m"
    BRIGHT_CYAN = "\033[96m"
    BRIGHT_WHITE = "\033[97m"
    END = "\033[0m"


class Level(Enum):
    """An exit message prefix."""

    CRITICAL = "Critical"
    DEBUG = "Debug"
    ERROR = "Error"
    HINT = "Hint"
    INFO = "Info"
    NOTE = "Note"
    WARNING = "Warning"

    @property
    def log_level(self: Level) -> int:
        """Return a log level.

        :returns: The log level
        """
        mapping = {
            Level.CRITICAL: logging.CRITICAL,
            Level.DEBUG: logging.DEBUG,
            Level.ERROR: logging.ERROR,
            Level.HINT: logging.INFO,
            Level.INFO: logging.INFO,
            Level.NOTE: logging.INFO,
            Level.WARNING: logging.WARNING,
        }
        return mapping[self]

    @classmethod
    def _longest_name(cls: type[T]) -> int:
        """Return the longest exit message prefix.

        :returns: The longest exit message prefix
        """
        return max(len(member.value) for member in cls)

    @classmethod
    def longest_formatted(cls: type[T]) -> int:
        """Return the longest exit message prefix.

        :returns: The longest exit message prefix
        """
        return max(len(str(member)) for member in cls)

    def __str__(self: Level) -> str:
        """Return the exit message prefix as a string.

        :returns: The exit message prefix as a string
        """
        return (
            f"{' ' * (self._longest_name() - len(self.name))}{self.name.capitalize()}: "
        )


@dataclass
class Msg:
    """An object to hold a message to present when exiting."""

    #: The message that will be presented
    message: str
    #: The prefix for the message, used for formatting
    prefix: Level = Level.ERROR

    @property
    def color(self: Msg) -> str:
        """Return a color for the prefix.

        :returns: The color for the prefix
        """
        color_mapping = {
            Level.CRITICAL: Color.BRIGHT_RED,
            Level.DEBUG: Color.GREY,
            Level.ERROR: Color.RED,
            Level.HINT: Color.CYAN,
            Level.INFO: Color.MAGENTA,
            Level.NOTE: Color.GREEN,
            Level.WARNING: Color.YELLOW,
        }
        return color_mapping[self.prefix]

    def to_lines(
        self: Msg,
        color: bool,  # noqa: FBT001
        width: int,
        with_prefix: bool,  # noqa: FBT001
    ) -> list[str]:
        """Output exit message to the console.

        :param color: Whether to color the message
        :param width: Constrain message to width
        :param with_prefix: Whether to prefix the message
        :returns: The exit message as a string
        """
        prefix_length = Level.longest_formatted()
        indent = " " * prefix_length

        lines = []
        message_lines = self.message.splitlines()

        lines.extend(
            textwrap.fill(
                message_lines[0],
                width=width,
                break_on_hyphens=False,
                initial_indent=str(self.prefix) if with_prefix else indent,
                subsequent_indent=indent,
            ).splitlines(),
        )

        if len(message_lines) > 1:
            for line in message_lines[1:]:
                lines.extend(
                    textwrap.fill(
                        line,
                        width=width,
                        break_on_hyphens=False,
                        initial_indent=indent,
                        subsequent_indent=indent,
                    ).splitlines(),
                )

        start_color = self.color if color else ""
        end_color = Color.END if color else ""

        return [f"{start_color}{line}{end_color}" for line in lines]


class Output:
    """Output functionality."""

    def __init__(  # noqa: PLR0913
        self: Output,
        log_file: str,
        log_level: str,
        log_append: str,
        term_features: TermFeatures,
        verbosity: int,
    ) -> None:
        """Initialize the output object.

        Args:
            log_file: The path to the los.get_terminal_size()og file
            log_level: The log level
            log_append: Whether to append to the log file
            term_features: Terminal features
            verbosity: The verbosity level
        """
        self._verbosity = verbosity
        self.call_count: dict[str, int] = {
            "critical": 0,
            "debug": 0,
            "error": 0,
            "hint": 0,
            "info": 0,
            "note": 0,
            "warning": 0,
        }
        self.term_features = term_features
        self.logger = logging.getLogger("pip4a")
        if log_level != "notset":
            self.logger.setLevel(log_level.upper())
            self.log_to_file = bool(log_file)
            log_file_path = Path(log_file)
            if log_file_path.exists() and log_append == "false":
                log_file_path.unlink()
            formatter = logging.Formatter(
                fmt="%(asctime)s %(levelname)s '%(name)s.%(module)s.%(funcName)s' %(message)s",
            )
            handler = logging.FileHandler(log_file)
            handler.setFormatter(formatter)
            handler.setLevel(log_level.upper())
            self.logger.addHandler(handler)
            self.log_to_file = True
        else:
            self.log_to_file = False

    def critical(self: Output, msg: str) -> None:
        """Print a critical message to the console.

        :param msg: The message to print
        """
        self.call_count["critical"] += 1
        self.log(msg, level=Level.CRITICAL)
        sys.exit(1)

    def debug(self: Output, msg: str) -> None:
        """Print a debug message to the console.

        :param msg: The message to print
        """
        self.call_count["debug"] += 1
        self.log(msg, level=Level.DEBUG)

    def error(self: Output, msg: str) -> None:
        """Print an error message to the console.

        :param msg: The message to print
        """
        self.call_count["error"] += 1
        self.log(msg, level=Level.ERROR)

    def hint(self: Output, msg: str) -> None:
        """Print a hint message to the console.

        :param msg: The message to print
        """
        self.call_count["hint"] += 1
        self.log(msg, level=Level.HINT)

    def info(self: Output, msg: str) -> None:
        """Print a hint message to the console.

        :param msg: The message to print
        """
        self.call_count["info"] += 1
        self.log(msg, level=Level.INFO)

    def note(self: Output, msg: str) -> None:
        """Print a note message to the console.

        :param msg: The message to print
        """
        self.call_count["note"] += 1
        self.log(msg, level=Level.NOTE)

    def warning(self: Output, msg: str) -> None:
        """Print a warning message to the console.

        :param msg: The message to print
        """
        self.call_count["warning"] += 1
        self.log(msg, level=Level.WARNING)

    def log(self: Output, msg: str, level: Level = Level.ERROR) -> None:
        """Print a message to the console.

        :param msg: The message to print
        :param prefix: The prefix for the message
        """
        if self.log_to_file:
            self.logger.log(level.log_level, msg, stacklevel=3)

        try:
            width = float(os.get_terminal_size()[0])
        except OSError:
            width = 80
        narrow = 80
        wide = 120
        if width <= narrow:
            set_width = width - 2
        elif wide >= width >= narrow:
            set_width = width * 0.9
        elif width > wide:
            set_width = width * 0.8

        debug = 2
        info = 1
        if (self._verbosity < debug and level == Level.DEBUG) or (
            self._verbosity < info and level == Level.INFO
        ):
            return

        lines = Msg(message=msg, prefix=level).to_lines(
            color=self.term_features.color,
            width=int(set_width),
            with_prefix=True,
        )
        if level in (Level.CRITICAL, Level.ERROR):
            print("\n".join(lines), file=sys.stderr)  # noqa: T201
        else:
            print("\n".join(lines))  # noqa: T201



"""A collection abstraction."""
from __future__ import annotations

import re
import sys

from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import yaml


if TYPE_CHECKING:
    from .config import Config
    from .output import Output


@dataclass
class Collection:  # pylint: disable=too-many-instance-attributes
    """A collection request specification."""

    config: Config
    path: Path | None = None
    opt_deps: str | None = None
    local: bool | None = None
    cnamespace: str | None = None
    cname: str | None = None
    specifier: str | None = None
    original: str | None = None

    @property
    def name(self: Collection) -> str:
        """Return the collection name."""
        return f"{self.cnamespace}.{self.cname}"

    @property
    def cache_dir(self: Collection) -> Path:
        """Return the collection cache directory."""
        collection_cache_dir = self.config.venv_cache_dir / self.name
        if not collection_cache_dir.exists():
            collection_cache_dir.mkdir()
        return collection_cache_dir

    @property
    def build_dir(self: Collection) -> Path:
        """Return the collection cache directory."""
        collection_build_dir = self.cache_dir / "build"
        if not collection_build_dir.exists():
            collection_build_dir.mkdir()
        return collection_build_dir

    @property
    def site_pkg_path(self: Collection) -> Path:
        """Return the site packages collection path."""
        if not self.cnamespace or not self.cname:
            msg = "Collection namespace or name not set."
            raise RuntimeError(msg)
        return self.config.site_pkg_collections_path / self.cnamespace / self.cname


def parse_collection_request(  # noqa: PLR0915
    string: str,
    config: Config,
    output: Output,
) -> Collection:
    """Parse a collection request str.

    Args:
        string: The collection request string
        config: The configuration object
        output: The output object
    Returns:
        A collection object
    """
    collection = Collection(config=config, original=string)
    # spec with dep, local
    if "[" in string and "]" in string:
        msg = f"Found optional dependencies in collection request: {string}"
        output.debug(msg)
        path = Path(string.split("[")[0]).expanduser().resolve()
        if not path.exists():
            msg = "Provide an existing path to a collection when specifying optional dependencies."
            output.hint(msg)
            msg = f"Failed to find collection path: {path}"
            output.critical(msg)
        msg = f"Found local collection request with dependencies: {string}"
        output.debug(msg)
        collection.path = path
        msg = f"Setting collection path: {collection.path}"
        collection.opt_deps = string.split("[")[1].split("]")[0]
        msg = f"Setting optional dependencies: {collection.opt_deps}"
        output.debug(msg)
        collection.local = True
        msg = "Setting request as local"
        output.debug(msg)
        get_galaxy(collection=collection, output=output)
        return collection
    # spec without dep, local
    path = Path(string).expanduser().resolve()
    if path.exists():
        msg = f"Found local collection request without dependencies: {string}"
        output.debug(msg)
        msg = f"Setting collection path: {path}"
        output.debug(msg)
        collection.path = path
        msg = "Setting request as local"
        output.debug(msg)
        collection.local = True
        get_galaxy(collection=collection, output=output)
        return collection
    non_local_re = re.compile(
        r"""
        (?P<cnamespace>[A-Za-z0-9]+)    # collection name
        \.                              # dot
        (?P<cname>[A-Za-z0-9]+)         # collection name
        (?P<specifier>[^A-Za-z0-9].*)?   # optional specifier
        """,
        re.VERBOSE,
    )
    matched = non_local_re.match(string)
    if not matched:
        msg = (
            "Specify a valid collection name (ns.n) with an optional version specifier"
        )
        output.hint(msg)
        msg = f"Failed to parse collection request: {string}"
        output.critical(msg)
        sys.exit(1)
    msg = f"Found non-local collection request: {string}"
    output.debug(msg)

    collection.cnamespace = matched.group("cnamespace")
    msg = f"Setting collection namespace: {collection.cnamespace}"
    output.debug(msg)

    collection.cname = matched.group("cname")
    msg = f"Setting collection name: {collection.cname}"
    output.debug(msg)

    if matched.group("specifier"):
        collection.specifier = matched.group("specifier")
        msg = f"Setting collection specifier: {collection.specifier}"
        output.debug(msg)

    collection.local = False
    msg = "Setting request as non-local"
    output.debug(msg)

    return collection


def get_galaxy(collection: Collection, output: Output) -> None:
    """Retrieve the collection name from the galaxy.yml file.

    Args:
        collection: A collection object
        output: The output object
    Raises:
        SystemExit: If the collection name is not found
    """
    if collection is None or collection.path is None:
        msg = "get_galaxy called without a collection or path"
        raise RuntimeError(msg)
    file_name = collection.path / "galaxy.yml"
    if not file_name.exists():
        err = f"Failed to find {file_name} in {collection.path}"
        output.critical(err)

    with file_name.open(encoding="utf-8") as fileh:
        try:
            yaml_file = yaml.safe_load(fileh)
        except yaml.YAMLError as exc:
            err = f"Failed to load yaml file: {exc}"
            output.critical(err)

    try:
        collection.cnamespace = yaml_file["namespace"]
        collection.cname = yaml_file["name"]
        msg = f"Found collection name: {collection.name} from {file_name}."
        output.debug(msg)
    except KeyError as exc:
        err = f"Failed to find collection name in {file_name}: {exc}"
        output.critical(err)
    else:
        return
    raise SystemExit(1)  # We shouldn't be here



"""The dependency checker."""

from __future__ import annotations

import json
import subprocess

from typing import TYPE_CHECKING

from packaging.specifiers import InvalidSpecifier, SpecifierSet
from packaging.version import Version

from pip4a.utils import (
    builder_introspect,
    collect_manifests,
    oxford_join,
    subprocess_run,
)


if TYPE_CHECKING:
    from pip4a.config import Config
    from pip4a.output import Output


class Checker:
    """The dependency checker."""

    def __init__(self: Checker, config: Config, output: Output) -> None:
        """Initialize the checker.

        Args:
            config: The application configuration.
            output: The application output object.
        """
        self._config: Config = config
        self._collections_missing: bool
        self._output: Output = output
        self._system_dep_missing: bool

    def run(self: Checker) -> None:
        """Run the checker."""
        builder_introspect(config=self._config, output=self._output)
        self._collection_deps()
        self.system_deps()
        self._python_deps()

    def _collection_deps(self: Checker) -> None:  # noqa: C901, PLR0912, PLR0915
        """Check collection dependencies."""
        collections = collect_manifests(
            target=self._config.site_pkg_collections_path,
            venv_cache_dir=self._config.venv_cache_dir,
        )
        missing = False
        for collection_name, details in collections.items():
            error = "Collection {collection_name} has malformed metadata."
            if not isinstance(details, dict):
                self._output.error(error)
                continue
            if not isinstance(details["collection_info"], dict):
                self._output.error(error)
                continue
            if not isinstance(details["collection_info"]["dependencies"], dict):
                self._output.error(error)
                continue

            msg = f"Checking dependencies for {collection_name}."
            self._output.debug(msg)

            deps = details["collection_info"]["dependencies"]

            if not deps:
                msg = f"Collection {collection_name} has no dependencies."
                self._output.debug(msg)
                continue
            for dep, version in deps.items():
                if not isinstance(version, str):
                    err = (
                        f"Collection {collection_name} has malformed"
                        f" dependency version for {dep}."
                    )
                    self._output.error(err)
                    continue
                try:
                    spec = SpecifierSet(version)
                except InvalidSpecifier:
                    spec = SpecifierSet(">=0.0.0")
                    msg = f"Invalid version specifier {version}, assuming >=0.0.0."
                    self._output.debug(msg)
                if dep in collections:
                    dependency = collections[dep]
                    error = "Collection {dep} has malformed metadata."
                    if not isinstance(dependency, dict):
                        self._output.error(error)
                        continue
                    if not isinstance(dependency["collection_info"], dict):
                        self._output.error(error)
                        continue

                    dep_version = dependency["collection_info"]["version"]
                    if not isinstance(dep_version, str):
                        self._output.error(error)
                        continue
                    dep_spec = Version(dep_version)
                    if not spec.contains(dep_spec):
                        err = (
                            f"Collection {collection_name} requires {dep} {version}"
                            f" but {dep} {dep_version} is installed."
                        )
                        self._output.error(err)
                        missing = True

                    else:
                        msg = (
                            f"Collection {collection_name} requires {dep} {version}"
                            f" and {dep} {dep_version} is installed."
                        )
                        self._output.debug(msg)
                else:
                    err = (
                        f"Collection {collection_name} requires"
                        f" {dep} {version} but it is not installed."
                    )
                    self._output.error(err)
                    msg = f"Try running `pip4a install {dep}`"
                    self._output.hint(msg)
                    missing = True

        if not missing:
            msg = "All dependant collections are installed."
            self._output.note(msg)
        self._collections_missing = missing

    def _python_deps(self: Checker) -> None:
        """Check Python dependencies."""
        if self._system_dep_missing:
            msg = "System packages are missing. Python dependency checking may fail."
            self._output.warning(msg)
            msg = "Install system packages and re-run `pip4a check`."
            self._output.hint(msg)
        missing_file = self._config.venv_cache_dir / "pip-report.txt"
        command = (
            f"{self._config.venv_interpreter} -m pip install -r"
            f" {self._config.discovered_python_reqs} --dry-run"
            f" --report {missing_file}"
        )
        work = "Building python package dependency tree"

        try:
            subprocess_run(
                command=command,
                verbose=self._config.args.verbose,
                msg=work,
                output=self._output,
            )
        except subprocess.CalledProcessError as exc:
            err = f"Failed to check python dependencies: {exc}"
            self._output.critical(err)
        with missing_file.open() as file:
            pip_report = json.load(file)

        if self._collections_missing:
            msg = "Python packages required by missing collections are not included."
            self._output.note(msg)

        if "install" not in pip_report or not pip_report["install"]:
            if not self._collections_missing:
                msg = "All python dependencies are installed."
                self._output.note(msg)
            return

        missing = [
            f"{package['metadata']['name']}=={package['metadata']['version']}"
            for package in pip_report["install"]
        ]

        err = f"Missing python dependencies: {oxford_join(missing)}"
        self._output.error(err)
        msg = f"Try running `pip install {' '.join(missing)}`."
        self._output.hint(msg)

    def system_deps(self: Checker) -> None:
        """Check the bindep file."""
        msg = "Checking system packages."
        self._output.info(msg)

        command = f"bindep -b -f {self._config.discovered_bindep_reqs}"
        work = "Checking system package requirements"
        try:
            subprocess_run(
                command=command,
                verbose=self._config.args.verbose,
                msg=work,
                output=self._output,
            )
        except subprocess.CalledProcessError as exc:
            if exc.stderr:
                msg = f"Bindep failed to find required system packages. {exc.stderr}"
                self._output.error(msg)
                msg = "Check the format of the bindep.txt file."
                self._output.hint(msg)
                self._system_dep_missing = True
                return
            lines = exc.stdout.splitlines()
            msg = (
                "Required system packages are missing."
                " Please use the system package manager to install them."
                "\n- "
            )
            msg += "\n- ".join(lines)
            self._output.warning(msg)
            self._system_dep_missing = True
        else:
            msg = "All required system packages are installed."
            self._output.note(msg)
            self._system_dep_missing = False
            return



"""Generate a dependency tree."""


from __future__ import annotations

from typing import TYPE_CHECKING, Union

from pip4a.tree import Tree
from pip4a.utils import builder_introspect, collect_manifests


if TYPE_CHECKING:
    from pip4a.config import Config
    from pip4a.output import Output

ScalarVal = Union[bool, str, float, int, None]
JSONVal = Union[ScalarVal, list["JSONVal"], dict[str, "JSONVal"]]


class TreeMaker:
    """Generate a dependency tree."""

    def __init__(self: TreeMaker, config: Config, output: Output) -> None:
        """Initialize the object.

        Args:
            config: The application configuration.
            output: The application output object.
        """
        self._config = config
        self._output = output

    def run(self: TreeMaker) -> None:  # noqa: C901, PLR0912, PLR0915
        """Run the command."""
        # pylint: disable=too-many-locals
        builder_introspect(self._config, self._output)

        with self._config.discovered_python_reqs.open("r") as reqs_file:
            python_deps = reqs_file.read().splitlines()

        collections = collect_manifests(
            target=self._config.site_pkg_collections_path,
            venv_cache_dir=self._config.venv_cache_dir,
        )
        tree_dict: JSONVal = {c: {} for c in collections}
        if not isinstance(tree_dict, dict):
            msg = "Tree dict is not a dict."
            raise TypeError(msg)

        links: dict[str, str] = {}
        for collection_name, collection in collections.items():
            err = f"Collection {collection_name} has malformed metadata."
            if not isinstance(collection["collection_info"], dict):
                self._output.error(err)
                continue
            if not isinstance(collection["collection_info"]["dependencies"], dict):
                self._output.error(err)
                continue

            for dep in collection["collection_info"]["dependencies"]:
                if not isinstance(dep, str):
                    err = f"Collection {collection_name} has malformed dependency."
                    self._output.error(err)
                    continue
                target = tree_dict[collection_name]
                if not isinstance(target, dict):
                    msg = "Tree dict is not a dict."
                    raise TypeError(msg)

                target[dep] = tree_dict[dep]

            docs = collection["collection_info"].get("documentation")
            homepage = collection["collection_info"].get("homepage")
            repository = collection["collection_info"].get("repository")
            issues = collection["collection_info"].get("issues")
            link = repository or homepage or docs or issues or "http://ansible.com"
            if not isinstance(link, str):
                msg = "Link is not a string."
                raise TypeError(msg)
            links[collection_name] = link

            if self._config.args.verbose >= 1:
                add_python_reqs(
                    tree_dict=tree_dict,
                    collection_name=collection_name,
                    python_deps=python_deps,
                )
        green: list[str] = []
        if self._config.args.verbose >= 1:
            green.append("python requirements")
            for line in python_deps:
                if "#" not in line:
                    green.append(line.strip())
                green.append(line.split("#", 1)[0].strip())

        more_verbose = 2
        if self._config.args.verbose >= more_verbose:
            tree = Tree(obj=tree_dict, term_features=self._config.term_features)
            tree.links = links
            tree.green.extend(green)
            rendered = tree.render()
            print(rendered)  # noqa: T201
        else:
            pruned_tree_dict: JSONVal = {}
            if not isinstance(pruned_tree_dict, dict):
                msg = "Tree dict is not a dict."
                raise TypeError(msg)
            for collection_name in list(tree_dict.keys()):
                found = False
                for value in tree_dict.values():
                    if not isinstance(value, dict):
                        msg = "Tree dict is not a dict."
                        raise TypeError(msg)
                    if collection_name in value:
                        found = True
                if not found:
                    pruned_tree_dict[collection_name] = tree_dict[collection_name]

            tree = Tree(obj=pruned_tree_dict, term_features=self._config.term_features)
            tree.links = links
            tree.green.extend(green)
            rendered = tree.render()
            print(rendered)  # noqa: T201

        if self._config.args.verbose >= 1:
            msg = "Only direct python dependencies are shown."
            self._output.info(msg)
            hint = "Run `pip show <pkg>` to see indirect dependencies."
            self._output.hint(hint)


def add_python_reqs(
    tree_dict: dict[str, JSONVal],
    collection_name: str,
    python_deps: list[str],
) -> None:
    """Add Python dependencies to the tree.

    Args:
        tree_dict: The tree dict.
        collection_name: The collection name.
        python_deps: The Python dependencies.

    Raises:
        TypeError: If the tree dict is not a dict.
    """
    if not isinstance(tree_dict, dict):
        msg = "Tree dict is not a dict."
        raise TypeError(msg)
    collection = tree_dict[collection_name]
    if not isinstance(collection, dict):
        msg = "Tree dict is not a dict."
        raise TypeError(msg)
    collection["python requirements"] = []

    for dep in sorted(python_deps):
        name, comment = dep.split("#", 1)
        if collection_name in comment:
            if not isinstance(collection["python requirements"], list):
                msg = "Python requirements is not a list."
                raise TypeError(msg)
            collection["python requirements"].append(name.strip())



"""The installer."""

from __future__ import annotations

import shutil

from pathlib import Path
from typing import TYPE_CHECKING

from pip4a.collection import Collection, parse_collection_request
from pip4a.utils import collections_from_requirements


if TYPE_CHECKING:
    from pip4a.config import Config
    from pip4a.output import Output


class UnInstaller:
    """The uninstaller class."""

    def __init__(self: UnInstaller, config: Config, output: Output) -> None:
        """Initialize the installer.

        Args:
            config: The application configuration.
            output: The application output object.
        """
        self._config = config
        self._output = output
        self._collection: Collection

    def run(self: UnInstaller) -> None:
        """Run the uninstaller."""
        if len(self._config.args.collection_specifier) > 1:
            msg = "Only one collection can be uninstalled at a time."
            self._output.critical(msg)

        if self._config.args.requirement:
            requirements_path = Path(self._config.args.requirement)
            if not requirements_path.exists():
                err = f"Failed to find requirements file: {requirements_path}"
                self._output.critical(err)
            collections = collections_from_requirements(requirements_path)
            for collection in collections:
                self._collection = parse_collection_request(
                    string=collection["name"],
                    config=self._config,
                    output=self._output,
                )
                self._remove_collection()
        else:
            self._collection = parse_collection_request(
                string=self._config.args.collection_specifier[0],
                config=self._config,
                output=self._output,
            )
            self._remove_collection()

    def _remove_collection(self: UnInstaller) -> None:
        """Remove the collection."""
        msg = f"Checking {self._collection.name} at {self._collection.site_pkg_path}"
        self._output.debug(msg)

        if self._collection.site_pkg_path.exists():
            msg = f"Exists: {self._collection.site_pkg_path}"
            self._output.debug(msg)

            if self._collection.site_pkg_path.is_symlink():
                self._collection.site_pkg_path.unlink()
            else:
                shutil.rmtree(self._collection.site_pkg_path)
            msg = f"Removed {self._collection.name}"
            self._output.note(msg)
        else:
            err = (
                f"Failed to find {self._collection.name}:"
                f" {self._collection.site_pkg_path}"
            )
            self._output.warning(err)

        for entry in self._config.site_pkg_collections_path.iterdir():
            if all(
                (
                    entry.is_dir(),
                    entry.name.startswith(self._collection.name),
                    entry.suffix == ".info",
                ),
            ):
                shutil.rmtree(entry)
                msg = f"Removed {self._collection.name}*.info: {entry}"
                self._output.debug(msg)

        collection_namespace_root = self._collection.site_pkg_path.parent
        try:
            collection_namespace_root.rmdir()
            msg = f"Removed collection namespace root: {collection_namespace_root}"
            self._output.debug(msg)
        except FileNotFoundError:
            pass
        except OSError as exc:
            msg = f"Failed to remove collection namespace root: {exc}"
            self._output.debug(msg)

        try:
            self._config.site_pkg_collections_path.rmdir()
            msg = f"Removed collection root: {self._config.site_pkg_collections_path}"
            self._output.debug(msg)
        except FileNotFoundError:
            pass
        except OSError as exc:
            msg = f"Failed to remove collection root: {exc}"
            self._output.debug(msg)



"""Unit test for the utilities module."""

from __future__ import annotations

from argparse import Namespace
from pathlib import Path

import pytest

from pip4a.collection import Collection, parse_collection_request
from pip4a.config import Config
from pip4a.output import Output
from pip4a.utils import TermFeatures


term_features = TermFeatures(color=False, links=False)

output = Output(
    log_file=str(Path.cwd() / "pip4a.log"),
    log_level="notset",
    log_append="true",
    term_features=term_features,
    verbosity=0,
)
config = Config(
    args=Namespace(),
    term_features=term_features,
    output=output,
)

FIXTURE_DIR = Path(__file__).parent.parent.resolve() / "fixtures"
scenarios = (
    (
        "ansible.utils",
        Collection(
            config=config,
            cname="utils",
            cnamespace="ansible",
            local=False,
            original="ansible.utils",
        ),
    ),
    (
        "ansible.utils:1.0.0",
        Collection(
            config=config,
            cname="utils",
            cnamespace="ansible",
            specifier=":1.0.0",
            local=False,
            original="ansible.utils:1.0.0",
        ),
    ),
    (
        "ansible.utils>=1.0.0",
        Collection(
            config=config,
            cname="utils",
            cnamespace="ansible",
            specifier=">=1.0.0",
            local=False,
            original="ansible.utils>=1.0.0",
        ),
    ),
    (
        str(FIXTURE_DIR),
        Collection(
            cname="cname",
            cnamespace="cnamespace",
            config=config,
            local=True,
            path=FIXTURE_DIR,
            specifier=None,
            original=str(FIXTURE_DIR),
        ),
    ),
    (
        str(FIXTURE_DIR) + "/[test]",
        Collection(
            cname="cname",
            cnamespace="cnamespace",
            config=config,
            local=True,
            opt_deps="test",
            path=FIXTURE_DIR,
            specifier=None,
            original=str(FIXTURE_DIR) + "/[test]",
        ),
    ),
    (
        "/foo/bar",
        None,
    ),
    (
        "abcdefg",
        None,
    ),
)


@pytest.mark.parametrize("scenario", scenarios)
def test_parse_collection_request(scenario: tuple[str, Collection | None]) -> None:
    """Test that the parse_collection_request function works as expected."""
    string, spec = scenario
    if spec is None:
        with pytest.raises(SystemExit):
            parse_collection_request(string=string, config=config, output=output)
    else:
        assert (
            parse_collection_request(string=string, config=config, output=output)
            == spec
        )



"""Lister module for pip4a."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pip4a.utils import collect_manifests, term_link


if TYPE_CHECKING:
    from pip4a.config import Config
    from pip4a.output import Output


class Lister:
    """The Lister class."""

    def __init__(self: Lister, config: Config, output: Output) -> None:
        """Initialize the Lister.

        Args:
            config: The application configuration.
            output: The application output object.
        """
        self._config = config
        self._output = output

    def run(self: Lister) -> None:
        """Run the Lister."""
        # pylint: disable=too-many-locals
        collections = collect_manifests(
            target=self._config.site_pkg_collections_path,
            venv_cache_dir=self._config.venv_cache_dir,
        )

        column1_width = 30
        column2_width = 10
        column3_width = 25

        print(  # noqa: T201
            f"{'Collection': <{column1_width}}"
            f" {'Version': <{column2_width}}"
            f" {'Editable project location': <{column3_width}}",
        )
        print(  # noqa: T201
            f"{'-' * (column1_width)}"
            f" {'-' * (column2_width)}"
            f" {'-' * (column3_width)}",
        )

        for fqcn, collection in collections.items():
            err = f"Collection {fqcn} has malformed metadata."
            ci = collection["collection_info"]
            if not isinstance(ci, dict):
                self._output.error(err)
                continue
            collection_name = ci["name"]
            collection_namespace = ci["namespace"]
            collection_version = ci["version"]
            if not isinstance(collection_name, str):
                self._output.error(err)
                continue
            if not isinstance(collection_namespace, str):
                self._output.error(err)
                continue
            if not isinstance(collection_version, str):
                self._output.error(err)
                continue

            collection_path = (
                self._config.site_pkg_collections_path
                / collection_namespace
                / collection_name
            )
            if collection_path.is_symlink():
                editable_location = str(collection_path.resolve())
            else:
                editable_location = ""

            docs = ci.get("documentation")
            homepage = ci.get("homepage")
            repository = ci.get("repository")
            issues = ci.get("issues")
            link = repository or homepage or docs or issues or "http://ansible.com"
            if not isinstance(link, str):
                msg = "Link is not a string."
                raise TypeError(msg)
            fqcn_linked = term_link(
                uri=link,
                label=fqcn,
                term_features=self._config.term_features,
            )

            print(  # noqa: T201
                fqcn_linked + " " * (column1_width - len(fqcn)),
                f"{ci['version']: <{column2_width}}",
                f"{editable_location: <{column3_width}}",
            )



"""CLI entrypoint."""

from __future__ import annotations

import os
import site
import sys

from pathlib import Path
from typing import TYPE_CHECKING

from pip4a import subcommands

from .arg_parser import parse
from .config import Config
from .output import Output
from .utils import TermFeatures


if TYPE_CHECKING:
    from argparse import Namespace


class Cli:
    """The Cli class."""

    def __init__(self: Cli) -> None:
        """Initialize the CLI and parse CLI args."""
        self.args: Namespace
        self.config: Config
        self.output: Output
        self.term_features: TermFeatures

    def parse_args(self: Cli) -> None:
        """Parse the command line arguments."""
        self.args = parse()
        if hasattr(self.args, "requirement") and self.args.requirement:
            self.args.requirement = Path(self.args.requirement).expanduser().resolve()

    def init_output(self: Cli) -> None:
        """Initialize the output object."""
        if not sys.stdout.isatty():
            self.term_features = TermFeatures(color=False, links=False)
        else:
            self.term_features = TermFeatures(
                color=False if os.environ.get("NO_COLOR") else not self.args.no_ansi,
                links=not self.args.no_ansi,
            )

        self.output = Output(
            log_append=self.args.log_append,
            log_file=self.args.log_file,
            log_level=self.args.log_level,
            term_features=self.term_features,
            verbosity=self.args.verbose,
        )

    def args_sanity(self: Cli) -> None:
        """Perform some sanity checking on the args."""
        # Missing args
        if (
            hasattr(self.args, "requirement")
            and self.args.requirement
            and not self.args.requirement.exists()
        ):
            err = f"Requirements file not found: {self.args.requirement}"
            self.output.critical(err)

        # Multiple editable collections
        if (
            hasattr(self.args, "collection_specifier")
            and len(self.args.collection_specifier) > 1
            and hasattr(self.args, "editable")
            and self.args.editable
        ):
            err = "Editable can only be used with a single collection specifier."
            self.output.critical(err)

        # Editable with requirements file
        if (
            hasattr(self.args, "requirement")
            and self.args.requirement
            and hasattr(self.args, "editable")
            and self.args.editable
        ):
            err = "Editable can not be used with a requirements file."
            self.output.critical(err)

    def ensure_isolated(self: Cli) -> None:
        """Ensure the environment is isolated."""
        env_vars = os.environ
        errored = False
        if "ANSIBLE_COLLECTIONS_PATHS" in env_vars:
            err = "ANSIBLE_COLLECTIONS_PATHS is set"
            self.output.error(err)
            hint = "Run `unset ANSIBLE_COLLECTIONS_PATHS` to unset it."
            self.output.hint(hint)
            errored = True
        if "ANSIBLE_COLLECTION_PATH" in env_vars:
            err = "ANSIBLE_COLLECTION_PATH is set"
            self.output.error(err)
            hint = "Run `unset ANSIBLE_COLLECTION_PATH` to unset it."
            self.output.hint(hint)
            errored = True

        home_coll = Path.home() / ".ansible/collections/ansible_collections"
        if home_coll.exists() and tuple(home_coll.iterdir()):
            err = f"Collections found in {home_coll}"
            self.output.error(err)
            hint = "Run `rm -rf ~/.ansible/collections` to remove them."
            self.output.hint(hint)
            errored = True

        usr_coll = Path("/usr/share/ansible/collections")
        if usr_coll.exists() and tuple(usr_coll.iterdir()):
            err = f"Collections found in {usr_coll}"
            self.output.error(err)
            hint = "Run `sudo rm -rf /usr/share/ansible/collections` to remove them."
            self.output.hint(hint)
            errored = True

        if "VIRTUAL_ENV" not in env_vars and not self.args.venv:
            err = (
                "Unable to use user site packages directory:"
                f" {site.getusersitepackages()}, please activate or specify a virtual environment"
            )
            self.output.error(err)
            hint = (
                "Use `--venv <directory>` to specify a virtual environment"
                " or enable an existing one."
            )
            self.output.hint(hint)
            errored = True

        if errored:
            err = (
                "The development environment is not isolated,"
                " please resolve the above errors."
            )

            self.output.critical(err)

    def run(self: Cli) -> None:
        """Run the application."""
        self.config = Config(
            args=self.args,
            output=self.output,
            term_features=self.term_features,
        )
        self.config.init()

        subcommand_cls = getattr(subcommands, self.config.args.subcommand.capitalize())
        subcommand = subcommand_cls(config=self.config, output=self.output)
        subcommand.run()
        self._exit()

    def _exit(self: Cli) -> None:
        """Exit the application setting the return code."""
        if self.output.call_count["error"]:
            sys.exit(1)
        if self.output.call_count["warning"]:
            sys.exit(2)
        sys.exit(0)


def main() -> None:
    """Entry point for ansible-creator CLI."""
    cli = Cli()
    cli.parse_args()
    cli.init_output()
    cli.args_sanity()
    cli.ensure_isolated()
    cli.run()


if __name__ == "__main__":
    main()



"""The subcommands module contains all the subcommands for pip4a."""

# ruff: noqa: F401

from .checker import Checker as Check
from .inspector import Inspector as Inspect
from .installer import Installer as Install
from .lister import Lister as List
from .treemaker import TreeMaker as Tree
from .uninstaller import UnInstaller as Uninstall



"""The collection inspect command."""

from __future__ import annotations

import json

from typing import TYPE_CHECKING

from pip4a.utils import collect_manifests


if TYPE_CHECKING:
    from pip4a.config import Config
    from pip4a.output import Output

try:
    from pip._vendor.rich import print_json

    HAS_RICH = True
except ImportError:
    HAS_RICH = False


class Inspector:
    """The Inspector class."""

    def __init__(self: Inspector, config: Config, output: Output) -> None:
        """Initialize the Inspector.

        Args:
            config: The application configuration.
            output: The application output object.
        """
        self._config = config
        self._output = output

    def run(self: Inspector) -> None:
        """Run the Inspector."""
        # pylint: disable=too-many-locals
        collections = collect_manifests(
            target=self._config.site_pkg_collections_path,
            venv_cache_dir=self._config.venv_cache_dir,
        )

        output = json.dumps(collections, indent=4, sort_keys=True)
        if HAS_RICH and self._config.term_features.color:
            print_json(output)
        else:
            print(output)  # noqa: T201



"""Parse the command line arguments."""

from __future__ import annotations

import argparse

from argparse import HelpFormatter
from pathlib import Path


try:
    from ._version import version as __version__
except ImportError:  # pragma: no cover
    try:
        import pkg_resources

        __version__ = pkg_resources.get_distribution("pip4a").version
    except Exception:  # pylint: disable=broad-except # noqa: BLE001
        # this is the fallback SemVer version picked by setuptools_scm when tag
        # information is not available.
        __version__ = "0.1.dev1"


def common_args(parser: ArgumentParser) -> None:
    """Add common arguments to the parser.

    Args:
        parser: The parser to add the arguments to
    """
    parser.add_argument(
        "--lf",
        "--log-file <file>",
        dest="log_file",
        default=str(Path.cwd() / "pip4a.log"),
        help="Log file to write to.",
    )
    parser.add_argument(
        "--ll",
        "--log-level <level>",
        dest="log_level",
        default="notset",
        choices=["notset", "debug", "info", "warning", "error", "critical"],
        help="Log level for file output.",
    )
    parser.add_argument(
        "--la",
        "--log-append <bool>",
        dest="log_append",
        choices=["true", "false"],
        default="true",
        help="Append to log file.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Give more CLI output. Option is additive, and can be used up to 3 times.",
    )


def parse() -> argparse.Namespace:
    """Parse the command line arguments.

    Returns:
        The arguments
    """
    parser = ArgumentParser(
        description="A pip-like ansible collection installer.",
        formatter_class=CustomHelpFormatter,
    )

    common_args(parser)

    parser.add_argument(
        "-V",
        "--version",
        action="version",
        help="Show version and exit.",
        version=__version__,
    )

    subparsers = parser.add_subparsers(
        title="Commands",
        dest="subcommand",
        metavar="",
        required=True,
    )

    level1 = ArgumentParser(add_help=False)

    level1.add_argument(
        "--ve",
        "--venv <directory>",
        help="Target virtual environment.",
        dest="venv",
    )

    level1.add_argument(
        "--na",
        "--no-ansi",
        action="store_true",
        default=False,
        dest="no_ansi",
        help="Disable the use of ANSI codes for terminal hyperlink generation and color.",
    )

    common_args(level1)

    _check = subparsers.add_parser(
        "check",
        formatter_class=CustomHelpFormatter,
        parents=[level1],
        help="Check installed collections",
    )

    _inspect = subparsers.add_parser(
        "inspect",
        formatter_class=CustomHelpFormatter,
        parents=[level1],
        help="Inspect installed collections",
    )

    _list = subparsers.add_parser(
        "list",
        formatter_class=CustomHelpFormatter,
        parents=[level1],
        help="List installed collections",
    )

    _tree = subparsers.add_parser(
        "tree",
        formatter_class=CustomHelpFormatter,
        parents=[level1],
        help="Generate a dependency tree",
    )

    level2 = ArgumentParser(add_help=False, parents=[level1])
    level2.add_argument(
        "collection_specifier",
        help="Collection name or path to collection with extras.",
        nargs="*",
    )
    level2.add_argument(
        "-r",
        "--requirement <file>",
        dest="requirement",
        help="Install from the given requirements file.",
        required=False,
    )

    install = subparsers.add_parser(
        "install",
        formatter_class=CustomHelpFormatter,
        parents=[level2],
        help="Install a collection.",
    )
    install.add_argument(
        "-e",
        "--editable",
        action="store_true",
        help="Install editable.",
    )

    _uninstall = subparsers.add_parser(
        "uninstall",
        formatter_class=CustomHelpFormatter,
        parents=[level2],
        help="Uninstall a collection.",
    )

    # pylint: disable=protected-access
    for grp in parser._action_groups:  # noqa: SLF001
        if grp.title is None:
            continue
        grp.title = grp.title.capitalize()
    for subparser in subparsers.choices.values():
        for grp in subparser._action_groups:  # noqa: SLF001
            if grp.title is None:
                continue
            grp.title = grp.title.capitalize()
    # pylint: enable=protected-access

    return parser.parse_args()


class ArgumentParser(argparse.ArgumentParser):
    """A custom argument parser."""

    def add_argument(  # type: ignore[no-untyped-def, override]
        self: ArgumentParser,
        *args,  # noqa: ANN002
        **kwargs,  # noqa: ANN003
    ) -> None:
        """Add an argument."""
        if "choices" in kwargs:
            kwargs["help"] += f" (choices: {', '.join(kwargs['choices'])})"
        if "default" in kwargs and kwargs["default"] != "==SUPPRESS==":
            kwargs["help"] += f" (default: {kwargs['default']})"
        kwargs["help"] = kwargs["help"][0].upper() + kwargs["help"][1:]
        super().add_argument(*args, **kwargs)


class CustomHelpFormatter(HelpFormatter):
    """A custom help formatter."""

    def __init__(self: CustomHelpFormatter, prog: str) -> None:
        """Initialize the help formatter.

        :param prog: The program name
        """
        long_string = "--abc  --really_really_really_log"
        # 3 here accounts for the spaces in the ljust(6) below
        HelpFormatter.__init__(
            self,
            prog=prog,
            indent_increment=1,
            max_help_position=len(long_string) + 3,
        )

    def _format_action_invocation(
        self: CustomHelpFormatter,
        action: argparse.Action,
    ) -> str:
        """Format the action invocation.

        Args:
            action: The action to format

        Raises:
            ValueError: If more than 2 options are given

        Returns:
            The formatted action invocation
        """
        if not action.option_strings:
            default = self._get_default_metavar_for_positional(action)
            (metavar,) = self._metavar_formatter(action, default)(1)
            return metavar

        if len(action.option_strings) == 1:
            return action.option_strings[0]

        max_variations = 2
        if len(action.option_strings) == max_variations:
            # Account for a --1234 --long-option-name
            return f"{action.option_strings[0].ljust(6)} {action.option_strings[1]}"
        msg = "Too many option strings"
        raise ValueError(msg)



"""Constants, for now, for pip4a."""

from __future__ import annotations

import json
import os
import subprocess
import sys

from pathlib import Path
from typing import TYPE_CHECKING

from .utils import subprocess_run


if TYPE_CHECKING:
    from argparse import Namespace

    from .output import Output
    from .utils import TermFeatures


class Config:
    """The application configuration."""

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self: Config,
        args: Namespace,
        output: Output,
        term_features: TermFeatures,
    ) -> None:
        """Initialize the configuration."""
        self._create_venv: bool = False
        self.args: Namespace = args
        self.bindir: Path
        self._output: Output = output
        self.python_path: Path
        self.site_pkg_path: Path
        self.venv_interpreter: Path
        self.term_features: TermFeatures = term_features

    def init(self: Config) -> None:
        """Initialize the configuration."""
        if self.args.venv:
            self._create_venv = True

        self._set_interpreter()
        self._set_site_pkg_path()

    @property
    def cache_dir(self: Config) -> Path:
        """Return the cache directory."""
        cache_dir = self.venv / ".pip4a"
        if not cache_dir.exists():
            cache_dir.mkdir(parents=True)
        return cache_dir

    @property
    def venv(self: Config) -> Path:
        """Return the virtual environment path."""
        if self.args.venv:
            return Path(self.args.venv).expanduser().resolve()
        venv_str = os.environ.get("VIRTUAL_ENV")
        if venv_str:
            return Path(venv_str).expanduser().resolve()
        err = "Failed to find a virtual environment."
        self._output.critical(err)
        sys.exit(1)

    @property
    def venv_cache_dir(self: Config) -> Path:
        """Return the virtual environment cache directory."""
        return self.cache_dir

    @property
    def discovered_python_reqs(self: Config) -> Path:
        """Return the discovered python requirements file."""
        return self.venv_cache_dir / "discovered_requirements.txt"

    @property
    def discovered_bindep_reqs(self: Config) -> Path:
        """Return the discovered system package requirements file."""
        return self.venv_cache_dir / "discovered_bindep.txt"

    @property
    def site_pkg_collections_path(self: Config) -> Path:
        """Return the site packages collection path."""
        site_pkg_collections_path = self.site_pkg_path / "ansible_collections"
        if not site_pkg_collections_path.exists():
            site_pkg_collections_path.mkdir()
        return site_pkg_collections_path

    @property
    def venv_bindir(self: Config) -> Path:
        """Return the virtual environment bin directory."""
        return self.venv / "bin"

    @property
    def interpreter(self: Config) -> Path:
        """Return the current interpreter."""
        return Path(sys.executable)

    def _set_interpreter(
        self: Config,
    ) -> None:
        """Set the interpreter."""
        if not self.venv.exists():
            if self._create_venv:
                msg = f"Creating virtual environment: {self.venv}"
                self._output.debug(msg)
                command = f"python -m venv {self.venv}"
                work = "Creating virtual environment"
                try:
                    subprocess_run(
                        command=command,
                        verbose=self.args.verbose,
                        msg=work,
                        output=self._output,
                    )
                    msg = f"Created virtual environment: {self.venv}"
                    self._output.info(msg)
                except subprocess.CalledProcessError as exc:
                    err = f"Failed to create virtual environment: {exc}"
                    self._output.critical(err)
            else:
                err = f"Cannot find virtual environment: {self.venv}."
                self._output.critical(err)
        msg = f"Virtual environment: {self.venv}"
        self._output.debug(msg)
        venv_interpreter = self.venv / "bin" / "python"
        if not venv_interpreter.exists():
            err = f"Cannot find interpreter: {venv_interpreter}."
            self._output.critical(err)

        msg = f"Virtual environment interpreter: {venv_interpreter}"
        self._output.debug(msg)
        self.venv_interpreter = venv_interpreter

    def _set_site_pkg_path(self: Config) -> None:
        """USe the interpreter to find the site packages path."""
        command = (
            f"{self.venv_interpreter} -c"
            " 'import json,site; print(json.dumps(site.getsitepackages()))'"
        )
        work = "Locating site packages directory"
        try:
            proc = subprocess_run(
                command=command,
                verbose=self.args.verbose,
                msg=work,
                output=self._output,
            )
        except subprocess.CalledProcessError as exc:
            err = f"Failed to find site packages path: {exc}"
            self._output.critical(err)

        try:
            site_pkg_dirs = json.loads(proc.stdout)
        except json.JSONDecodeError as exc:
            err = f"Failed to decode json: {exc}"
            self._output.critical(err)

        if not site_pkg_dirs:
            err = "Failed to find site packages path."
            self._output.critical(err)

        msg = f"Found site packages path: {site_pkg_dirs[0]}"
        self._output.debug(msg)

        self.site_pkg_path = Path(site_pkg_dirs[0])



# cspell:ignore mkey, mfour
"""Test the tree generator."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pip4a.tree import Tree
from pip4a.utils import TermFeatures


if TYPE_CHECKING:
    from pip4a.tree import JSONVal


sample_1: JSONVal = {
    "key_one": "one",
    "key_two": 42,
    "key_three": True,
    "key_four": None,
    "key_five": ["one", "two", "three"],
    "key_six": {
        "key_one": "one",
        "key_two": 42,
        "key_three": True,
        "key_four": None,
        "key_five": ["one", "two", "three"],
        "key_six": {
            "key_one": "one",
            "key_two": 42,
            "key_three": True,
            "key_four": None,
            "key_five": ["one", "two", "three"],
            "key_six": {
                "key_one": "one",
                "key_two": 42,
                "key_three": True,
                "key_four": None,
                "key_five": ["one", "two", "three"],
            },
        },
    },
    "key_seven": [{"foo": [1, 2, 3]}],
    "key_eight": [1, 2, 3],
}

result = r"""key_one
└──one
key_two
└──42
key_three
└──True
key_four
└──None
key_five
├──one
├──two
└──three
key_six
├──key_one
│  └──one
├──key_two
│  └──42
├──key_three
│  └──True
├──key_four
│  └──None
├──key_five
│  ├──one
│  ├──two
│  └──three
└──key_six
   ├──key_one
   │  └──one
   ├──key_two
   │  └──42
   ├──key_three
   │  └──True
   ├──key_four
   │  └──None
   ├──key_five
   │  ├──one
   │  ├──two
   │  └──three
   └──key_six
      ├──key_one
      │  └──one
      ├──key_two
      │  └──42
      ├──key_three
      │  └──True
      ├──key_four
      │  └──None
      └──key_five
         ├──one
         ├──two
         └──three
key_seven
└──foo
   ├──1
   ├──2
   └──3
key_eight
├──1
├──2
└──3
"""


def test_tree_large() -> None:
    """Test the tree generator."""
    term_features = TermFeatures(color=False, links=False)

    assert Tree(obj=sample_1, term_features=term_features).render() == result


sample_2: JSONVal = {
    "key_one": True,
    "key_two": 42,
    "key_three": None,
    "key_four": "four",
    "key_five": [{"a": 1}, {"b": 2}],
}

expected = [
    "\x1b[34mkey_one\x1b[0m",
    "└──\x1b[32mTrue\x1b[0m",
    "\x1b[34mkey_two\x1b[0m",
    "└──\x1b[32m42\x1b[0m",
    "\x1b[34mkey_three\x1b[0m",
    "└──\x1b[32mNone\x1b[0m",
    "\x1b[34mkey_four\x1b[0m",
    "└──\x1b[32mfour\x1b[0m",
    "key_five",
    "├──\x1b[3m0\x1b[0m",
    "│  └──a",
    "│     └──1",
    "└──\x1b[3m1\x1b[0m",
    "   └──b",
    "      └──\x1b[34m\x1b]8;;http://red.ht\x1b\\2\x1b]8;;\x1b\\\x1b[0m",
]


def test_tree_color() -> None:
    """Test the tree generator."""
    term_features = TermFeatures(color=True, links=True)
    tree = Tree(obj=sample_2, term_features=term_features)
    tree.blue = ["key_one", "key_two", "key_three", "key_four"]
    tree.green = [True, 42, None, "four"]
    tree.links = {"2": "http://red.ht"}
    rendered = tree.render().splitlines()
    assert rendered == expected



# file generated by setuptools_scm
# don't change, don't track in version control
__version__ = version = '0.2.17'  # type: str
__version_tuple__ = version_tuple = (0, 2, 17)  # type: tuple[int | str, ...]



"""The installer."""

from __future__ import annotations

import re
import shutil
import subprocess

from pathlib import Path
from typing import TYPE_CHECKING

from pip4a.collection import Collection, parse_collection_request
from pip4a.utils import (
    builder_introspect,
    collections_from_requirements,
    oxford_join,
    subprocess_run,
)

from .checker import Checker


if TYPE_CHECKING:
    from pip4a.config import Config
    from pip4a.output import Output


class Installer:
    """The installer class."""

    def __init__(self: Installer, config: Config, output: Output) -> None:
        """Initialize the installer.

        Args:
            config: The application configuration.
            output: The application output object.
        """
        self._config = config
        self._output = output
        self._current_collection_spec: str

    def run(self: Installer) -> None:
        """Run the installer."""
        if (
            self._config.args.collection_specifier
            and "," in self._config.args.collection_specifier
        ):
            err = "Multiple optional dependencies are not supported at this time."
            self._output.critical(err)

        self._install_core()

        if self._config.args.requirement:
            self._install_galaxy_requirements()
        if self._config.args.collection_specifier:
            collections = [
                parse_collection_request(
                    string=entry,
                    config=self._config,
                    output=self._output,
                )
                for entry in self._config.args.collection_specifier
            ]
            local_collections = [
                collection for collection in collections if collection.local
            ]
            for local_collection in local_collections:
                self._install_local_collection(collection=local_collection)
                if self._config.args.editable:
                    self._swap_editable_collection(collection=local_collection)
            distant_collections = [
                collection for collection in collections if not collection.local
            ]
            if distant_collections:
                if self._config.args.editable:
                    msg = "Editable installs are only supported for local collections."
                    self._output.critical(msg)
                self._install_galaxy_collections(collections=distant_collections)

        builder_introspect(config=self._config, output=self._output)
        self._pip_install()
        Checker(config=self._config, output=self._output).system_deps()

        if self._config.args.venv and (
            self._config.interpreter != self._config.venv_interpreter
        ):
            msg = "A virtual environment was specified but has not been activated."
            self._output.note(msg)
            msg = (
                "Please activate the virtual environment:"
                f"\nsource {self._config.args.venv}/bin/activate"
            )
            self._output.note(msg)

    def _install_core(self: Installer) -> None:
        """Install ansible-core if not installed already."""
        msg = "Installing ansible-core."
        self._output.info(msg)

        core = self._config.venv_bindir / "ansible"
        if core.exists():
            return
        msg = "Installing ansible-core."
        self._output.debug(msg)
        command = f"{self._config.venv_interpreter} -m pip install ansible-core"
        try:
            subprocess_run(
                command=command,
                verbose=self._config.args.verbose,
                msg=msg,
                output=self._output,
            )
        except subprocess.CalledProcessError as exc:
            err = f"Failed to install ansible-core: {exc}"
            self._output.critical(err)

    def _install_galaxy_collections(
        self: Installer,
        collections: list[Collection],
    ) -> None:
        """Install the collection from galaxy."""
        collections_str = " ".join(
            [f"'{collection.original}'" for collection in collections],
        )
        msg = f"Installing collections from galaxy: {collections_str}"
        self._output.info(msg)

        for collection in collections:
            if collection.site_pkg_path.exists():
                msg = f"Removing installed {collection.site_pkg_path}"
                self._output.debug(msg)
                if collection.site_pkg_path.is_symlink():
                    collection.site_pkg_path.unlink()
                else:
                    shutil.rmtree(collection.site_pkg_path)

        command = (
            f"{self._config.venv_bindir / 'ansible-galaxy'} collection"
            f" install {collections_str}"
            f" -p {self._config.site_pkg_path}"
            " --force"
        )
        env = {
            "ANSIBLE_GALAXY_COLLECTIONS_PATH_WARNING": str(self._config.args.verbose),
        }
        msg = "Running ansible-galaxy to install non-local collection and it's dependencies."
        self._output.debug(msg)
        try:
            proc = subprocess_run(
                command=command,
                env=env,
                verbose=self._config.args.verbose,
                msg=msg,
                output=self._output,
            )
        except subprocess.CalledProcessError as exc:
            err = f"Failed to install collection: {exc}\n{exc.stderr}"
            self._output.critical(err)
            return
        installed = re.findall(r"(\w+\.\w+):.*installed", proc.stdout)
        msg = f"Installed collections include: {oxford_join(installed)}"
        self._output.note(msg)

    def _install_galaxy_requirements(self: Installer) -> None:
        """Install the collections using requirements.yml."""
        msg = f"Installing collections from requirements file: {self._config.args.requirement}"
        self._output.info(msg)

        collections = collections_from_requirements(file=self._config.args.requirement)
        for collection in collections:
            cnamespace = collection["name"].split(".")[0]
            cname = collection["name"].split(".")[1]
            cpath = self._config.site_pkg_collections_path / cnamespace / cname
            if cpath.exists():
                msg = f"Removing installed {cpath}"
                self._output.debug(msg)
                if cpath.is_symlink():
                    cpath.unlink()
                else:
                    shutil.rmtree(cpath)

        command = (
            f"{self._config.venv_bindir / 'ansible-galaxy'} collection"
            f" install -r {self._config.args.requirement}"
            f" -p {self._config.site_pkg_path}"
            " --force"
        )
        work = "Install collections from requirements file"
        try:
            proc = subprocess_run(
                command=command,
                verbose=self._config.args.verbose,
                msg=work,
                output=self._output,
            )
        except subprocess.CalledProcessError as exc:
            err = f"Failed to install collections: {exc} {exc.stderr}"
            self._output.critical(err)

        installed = re.findall(r"(\w+\.\w+):.*installed", proc.stdout)
        msg = f"Installed collections include: {oxford_join(installed)}"
        self._output.note(msg)

    def _install_local_collection(  # noqa: PLR0915, PLR0912
        self: Installer,
        collection: Collection,
    ) -> None:
        """Install the collection from the build directory.

        Args:
            collection: The collection object.

        Raises:
            RuntimeError: If tarball is not found or if more than one tarball is found.
        """
        # pylint: disable=too-many-instance-attributes
        msg = f"Installing local collection from: {collection.build_dir}"
        self._output.info(msg)

        command = (
            "cp -r --parents $(git ls-files 2> /dev/null || ls)"
            f" {collection.build_dir}"
        )
        msg = "Copying collection to build directory using git ls-files."
        self._output.debug(msg)
        try:
            subprocess_run(
                command=command,
                cwd=collection.path,
                verbose=self._config.args.verbose,
                msg=msg,
                output=self._output,
            )
        except subprocess.CalledProcessError as exc:
            err = f"Failed to copy collection to build directory: {exc} {exc.stderr}"
            self._output.critical(err)

        command = (
            f"cd {collection.build_dir} &&"
            f" {self._config.venv_bindir / 'ansible-galaxy'} collection build"
            f" --output-path {collection.build_dir}"
            " --force"
        )

        msg = "Running ansible-galaxy to build collection."
        self._output.debug(msg)

        try:
            subprocess_run(
                command=command,
                verbose=self._config.args.verbose,
                msg=msg,
                output=self._output,
            )
        except subprocess.CalledProcessError as exc:
            err = f"Failed to build collection: {exc} {exc.stderr}"
            self._output.critical(err)

        built = [
            f
            for f in Path(collection.build_dir).iterdir()
            if f.is_file() and f.name.endswith(".tar.gz")
        ]
        if len(built) != 1:
            err = (
                "Expected to find one collection tarball in"
                f"{collection.build_dir}, found {len(built)}"
            )
            raise RuntimeError(err)
        tarball = built[0]

        if collection.site_pkg_path.exists():
            msg = f"Removing installed {collection.site_pkg_path}"
            self._output.debug(msg)
            if collection.site_pkg_path.is_symlink():
                collection.site_pkg_path.unlink()
            else:
                shutil.rmtree(collection.site_pkg_path)

        info_dirs = [
            entry
            for entry in self._config.site_pkg_collections_path.iterdir()
            if entry.is_dir()
            and entry.name.endswith(".info")
            and entry.name.startswith(collection.name)
        ]
        for info_dir in info_dirs:
            msg = f"Removing installed {info_dir}"
            self._output.debug(msg)
            shutil.rmtree(info_dir)

        command = (
            f"{self._config.venv_bindir / 'ansible-galaxy'} collection"
            f" install {tarball} -p {self._config.site_pkg_path}"
            " --force"
        )
        env = {
            "ANSIBLE_GALAXY_COLLECTIONS_PATH_WARNING": str(self._config.args.verbose),
        }
        msg = "Running ansible-galaxy to install a local collection and it's dependencies."
        self._output.debug(msg)
        try:
            proc = subprocess_run(
                command=command,
                env=env,
                verbose=self._config.args.verbose,
                msg=msg,
                output=self._output,
            )
        except subprocess.CalledProcessError as exc:
            err = f"Failed to install collection: {exc} {exc.stderr}"
            self._output.critical(err)
            return

        # ansible-galaxy collection install does not include the galaxy.yml for version
        # nor does it create an info file that can be used to determine the version.
        # preserve the MANIFEST.json file for editable installs
        if not self._config.args.editable:
            shutil.copy(
                collection.build_dir / "galaxy.yml",
                collection.site_pkg_path / "galaxy.yml",
            )
        else:
            shutil.copy(
                collection.site_pkg_path / "MANIFEST.json",
                collection.cache_dir / "MANIFEST.json",
            )

        installed = re.findall(r"(\w+\.\w+):.*installed", proc.stdout)
        msg = f"Installed collections include: {oxford_join(installed)}"
        self._output.note(msg)

    def _swap_editable_collection(self: Installer, collection: Collection) -> None:
        """Swap the installed collection with the current working directory.

        Args:
            collection: The collection object.

        Raises:
            RuntimeError: If the collection path is not set.
        """
        msg = f"Swapping {collection.name} with {collection.path}"
        self._output.info(msg)

        if collection.path is None:
            msg = "Collection path not set"
            raise RuntimeError(msg)
        msg = f"Removing installed {collection.site_pkg_path}"
        self._output.debug(msg)
        if collection.site_pkg_path.exists():
            if collection.site_pkg_path.is_symlink():
                collection.site_pkg_path.unlink()
            else:
                shutil.rmtree(collection.site_pkg_path)

        msg = f"Symlinking {collection.site_pkg_path} to {collection.path}"
        self._output.debug(msg)
        collection.site_pkg_path.symlink_to(collection.path)

    def _pip_install(self: Installer) -> None:
        """Install the dependencies."""
        msg = "Installing python requirements."
        self._output.info(msg)

        command = (
            f"{self._config.venv_interpreter} -m pip install"
            f" -r {self._config.discovered_python_reqs}"
        )

        msg = (
            f"Installing python requirements from {self._config.discovered_python_reqs}"
        )
        self._output.debug(msg)
        work = "Installing python requirements"
        try:
            subprocess_run(
                command=command,
                verbose=self._config.args.verbose,
                msg=work,
                output=self._output,
            )
        except subprocess.CalledProcessError as exc:
            err = (
                "Failed to install requirements from"
                f" {self._config.discovered_python_reqs}: {exc} {exc.stderr}"
            )
            self._output.critical(err)
        else:
            msg = "All python requirements are installed."
            self._output.note(msg)
